from datetime import date, timedelta

from .base import BaseModel, SourceMixin, AssociatedLinkMixin, LinkMixin
from .schemas.event import schema
from ..exceptions import ScrapeValueError
from ..utils import _make_pseudo_id, is_valid_uuid


def calculate_window(*, base_day=None, days_before=30, days_after=90):
    """given details on a window, returns start & end dates for windowing purposes"""
    if not base_day:
        base_day = date.today()
    start = base_day - timedelta(days=days_before)
    end = base_day + timedelta(days=days_after)
    return start, end


class EventAgendaItem(dict, AssociatedLinkMixin):
    event = None

    def __init__(self, description, event):
        super(EventAgendaItem, self).__init__(
            {
                "description": description,
                "classification": [],
                "related_entities": [],
                "subjects": [],
                "media": [],
                "notes": [],
                "order": len(event.agenda),
                "extras": {},
            }
        )
        self.event = event

    def add_subject(self, what):
        self["subjects"].append(what)

    def add_classification(self, what):
        self["classification"].append(what)

    def add_vote_event(self, vote_event, *, id=None, note="consideration"):
        self.add_entity(name=vote_event, entity_type="vote_event", id=id, note=note)

    def add_committee(self, committee, *, id=None, note="participant"):
        self.add_entity(name=committee, entity_type="organization", id=id, note=note)

    def add_bill(self, bill, *, id=None, note="consideration"):
        self.add_entity(name=bill, entity_type="bill", id=id, note=note)

    def add_person(self, person, *, id=None, note="participant"):
        self.add_entity(name=person, entity_type="person", id=id, note=note)

    def add_media_link(
        self,
        note,
        url,
        media_type,
        *,
        on_duplicate="warn",
        date="",
        classification="",
    ):
        return self._add_associated_link(
            collection="media",
            note=note,
            url=url,
            media_type=media_type,
            on_duplicate=on_duplicate,
            date=date,
            classification=classification,
        )

    def add_entity(self, name, entity_type, *, id, note):
        ret = {"name": name, "entity_type": entity_type, "note": note}
        if is_valid_uuid(id):
            ret["id"] = id
        elif entity_type:
            if entity_type in ("organization", "person"):
                id = _make_pseudo_id(name=name)
            elif entity_type in ("bill", "vote_event"):
                id = _make_pseudo_id(identifier=name)
            else:
                raise ScrapeValueError(
                    "attempt to call add_entity with unsupported "
                    "entity type: {}".format(entity_type)
                )
            ret[entity_type + "_id"] = id

        self["related_entities"].append(ret)


class Event(BaseModel, SourceMixin, AssociatedLinkMixin, LinkMixin):
    """
    Details for an event in .format
    """

    _type = "event"
    _schema = schema

    def __init__(
        self,
        name,
        start_date,
        location_name,
        *,
        all_day=False,
        description="",
        end_date="",
        status="confirmed",
        classification="event",
        upstream_id="",
    ):
        super(Event, self).__init__()
        self.start_date = start_date
        self.all_day = all_day
        self.end_date = end_date
        self.name = name
        self.description = description
        self.status = status
        self.classification = classification
        self.upstream_id = upstream_id
        self.location = {"name": location_name, "note": "", "coordinates": None}
        self.documents = []
        self.participants = []
        self.media = []
        self.agenda = []

    def __str__(self):
        return "{} {}".format(self.start_date, self.name.strip())

    def set_location(self, name, *, note="", url="", coordinates=None):
        self.location = {
            "name": name,
            "note": note,
            "url": url,
            "coordinates": coordinates,
        }

    def add_participant(self, name, type, *, id=None, note="participant"):
        p = {"name": name, "entity_type": type, "note": note}
        if is_valid_uuid(id):
            p["id"] = id
        elif type:
            id = _make_pseudo_id(name=name)
            p[type + "_id"] = id

        self.participants.append(p)

    def add_person(self, name, *, id=None, note="participant"):
        return self.add_participant(name=name, type="person", id=id, note=note)

    def add_committee(self, name, *, id=None, note="participant"):
        return self.add_participant(name=name, type="organization", id=id, note=note)

    def add_agenda_item(self, description):
        obj = EventAgendaItem(description, self)
        self.agenda.append(obj)
        return obj

    def add_bill(self, bill, *, note="consideration", agenda_item="Associated Bills"):
        """
        adds a dummy agenda item for associating bills for cases where we want bills
        but don't have appropriate agenda items

        context: https://github.com/openstates/enhancement-proposals/pull/28#issuecomment-898720989
        """
        for item in self.agenda:
            if item["description"] == agenda_item:
                break
        else:
            item = EventAgendaItem(agenda_item, self)
            self.agenda.append(item)
        item.add_bill(bill, note=note)

    def add_media_link(
        self,
        note,
        url,
        media_type,
        *,
        on_duplicate="error",
        date="",
        classification="",
    ):
        return self._add_associated_link(
            collection="media",
            note=note,
            url=url,
            media_type=media_type,
            on_duplicate=on_duplicate,
            date=date,
            classification=classification,
        )

    def add_document(
        self,
        note,
        url,
        *,
        media_type="",
        on_duplicate="error",
        date="",
        classification="",
    ):
        return self._add_associated_link(
            collection="documents",
            note=note,
            url=url,
            media_type=media_type,
            on_duplicate=on_duplicate,
            date=date,
            classification=classification,
        )



#!/usr/bin/env python
import os
import io
import hashlib
import click
import boto3  # type: ignore
import typing
from PIL import Image  # type: ignore
from botocore.exceptions import ClientError  # type: ignore
import requests
from .general import get_data_path
from ...models.people import Person


ALLOWED_CONTENT_TYPES = ("image/jpeg", "image/png", "image/gif", "image/jpg")


_IMAGE_RETURN_TYPE = tuple[typing.Optional[bytes], typing.Optional[str]]


def upload(
    img_callable: typing.Callable[[], _IMAGE_RETURN_TYPE],
    key_name: str,
    skip_existing: bool,
) -> typing.Optional[bytes]:
    """upload works as a sort of decorator around img_callable, which is
    only called if necessary after checking if there's already an image"""
    s3 = boto3.client("s3")
    try:
        obj = s3.head_object(Bucket=os.environ["S3_BUCKET"], Key=key_name)
    except ClientError:
        obj = None

    if obj and skip_existing:
        click.secho(f"{key_name} already exists", fg="yellow")
        return None

    # if we need to get the object, call the (potentially expensive) callable
    img_bytes, content_type = img_callable()
    if not img_bytes:
        return None

    # compare sha1 hashes
    sha1 = hashlib.sha1(img_bytes).hexdigest()
    if obj and obj["Metadata"].get("sha1") == sha1:
        click.secho(f"{key_name} already up to date", fg="yellow")
        return img_bytes

    click.secho(f"uploading {key_name}", fg="green")
    s3.upload_fileobj(
        io.BytesIO(img_bytes),
        os.environ["S3_BUCKET"],
        key_name,
        ExtraArgs={
            "Metadata": {"sha1": sha1},
            "ContentType": content_type,
            "ACL": "public-read",
        },
    )

    # return the raw bytes, which may be reused for resizing/etc.
    return img_bytes


def download_image(url: str) -> _IMAGE_RETURN_TYPE:
    try:
        resp = requests.get(url)
    except Exception as e:
        click.secho(f"could not fetch {url}, {e}", fg="red")
        return None, None

    if resp.status_code != 200:
        click.secho(f"could not fetch {url}, {resp.status_code}", fg="red")
        return None, None

    content_type = resp.headers["content-type"]
    if content_type not in ALLOWED_CONTENT_TYPES:
        click.secho(f"unknown content type for {url}, {content_type}", fg="red")
        return None, None

    return resp.content, content_type


def resize_image(img_bytes: bytes, size: int) -> _IMAGE_RETURN_TYPE:
    img = Image.open(fp=io.BytesIO(img_bytes))
    img = img.convert("RGB")  # type: ignore
    img.thumbnail((size, size))
    output = io.BytesIO()
    img.save(output, "JPEG", quality=80, progressive=True)
    output.seek(0)
    return output.read(), "image/jpeg"


def download_state_images(abbr: str, skip_existing: bool) -> None:
    for filename in (get_data_path(abbr) / "legislature").glob("*.yml"):
        person: Person = Person.load_yaml(filename)
        url = person.image
        if not url:
            continue

        # safe to cast because we've bailed above if url is None
        img_bytes = upload(
            lambda: download_image(typing.cast(str, url)),
            f"images/original/{person.id}",
            skip_existing,
        )
        # if the image got skipped, we can't do the resizes either, this means if we add new
        # profiles we need to run with --no-skip-existing
        if not img_bytes:
            continue

        # resize image so largest dimension is 200px
        upload(
            lambda: resize_image(typing.cast(bytes, img_bytes), 200),
            f"images/small/{person.id}",
            skip_existing,
        )


# def recognize(key):
#     client = boto3.client('rekognition')
#     resp = client.detect_faces(Image={'S3Object': {'Bucket': os.environ['S3_BUCKET'],
#                                                'Name': key}},
#                            Attributes=["DEFAULT"])
# algorithm suggested here seems like a good starting point
# https://stackoverflow.com/questions/4813608/cropping-an-image-with-a-focus-area-face-using-imagemagick



# Generated by Django 3.2.2 on 2021-09-20 23:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0039_auto_20210920_1934"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="eventdocumentlink",
            name="document",
        ),
        migrations.RemoveField(
            model_name="eventmedialink",
            name="media",
        ),
        migrations.AddField(
            model_name="eventagendamedia",
            name="links",
            field=models.JSONField(blank=True, default=list),
        ),
        migrations.AddField(
            model_name="eventdocument",
            name="links",
            field=models.JSONField(blank=True, default=list),
        ),
        migrations.AddField(
            model_name="eventmedia",
            name="links",
            field=models.JSONField(blank=True, default=list),
        ),
        migrations.DeleteModel(
            name="EventAgendaMediaLink",
        ),
        migrations.DeleteModel(
            name="EventDocumentLink",
        ),
        migrations.DeleteModel(
            name="EventMediaLink",
        ),
    ]



# Generated by Django 3.0.5 on 2021-02-03 19:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0028_auto_20201022_1642"),
    ]

    operations = [
        migrations.AlterField(
            model_name="eventagendaitem",
            name="order",
            field=models.PositiveIntegerField(),
        ),
    ]



from ..models import State, Chamber, District

DC = State(
    name="District of Columbia",
    abbr="DC",
    capital="None",
    capital_tz="America/New_York",
    fips="11",
    unicameral=True,
    legislature_name="Council of the District of Columbia",
    legislature_organization_id="ocd-organization/393504d5-b5ba-446f-908b-9868996bf3f7",
    executive_name="Office of the Governor",
    executive_organization_id="ocd-organization/86e53984-1898-47b4-b74b-b3de66455937",
    division_id="ocd-division/country:us/district:dc",
    jurisdiction_id="ocd-jurisdiction/country:us/district:dc/government",
    url="https://dc.gov",
    legislature=Chamber(
        chamber_type="unicameral",
        name="Council of the District of Columbia",
        organization_id="ocd-organization/393504d5-b5ba-446f-908b-9868996bf3f7",
        num_seats=13,
        title="Councilmember",
        districts=[
            District(
                "Ward 1", "legislature", "ocd-division/country:us/district:dc/ward:1", 1
            ),
            District(
                "Ward 2", "legislature", "ocd-division/country:us/district:dc/ward:2", 1
            ),
            District(
                "Ward 3", "legislature", "ocd-division/country:us/district:dc/ward:3", 1
            ),
            District(
                "Ward 4", "legislature", "ocd-division/country:us/district:dc/ward:4", 1
            ),
            District(
                "Ward 5", "legislature", "ocd-division/country:us/district:dc/ward:5", 1
            ),
            District(
                "Ward 6", "legislature", "ocd-division/country:us/district:dc/ward:6", 1
            ),
            District(
                "Ward 7", "legislature", "ocd-division/country:us/district:dc/ward:7", 1
            ),
            District(
                "Ward 8", "legislature", "ocd-division/country:us/district:dc/ward:8", 1
            ),
            District(
                "Chairman", "legislature", "ocd-division/country:us/district:dc", 1
            ),
            District(
                "At-Large", "legislature", "ocd-division/country:us/district:dc", 4
            ),
        ],
    ),
)



# Generated by Django 3.2.2 on 2021-09-20 19:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0038_auto_20210914_1559"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="eventsource",
            name="event",
        ),
        migrations.AddField(
            model_name="event",
            name="links",
            field=models.JSONField(blank=True, default=list),
        ),
        migrations.AddField(
            model_name="event",
            name="sources",
            field=models.JSONField(blank=True, default=list),
        ),
        migrations.DeleteModel(
            name="EventLink",
        ),
        migrations.DeleteModel(
            name="EventSource",
        ),
    ]



# Generated by Django 3.0.5 on 2020-08-04 15:24

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("data", "0011_auto_20200804_1108")]

    operations = [
        migrations.AddField(
            model_name="person",
            name="current_role",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                default=None, null=True
            ),
        )
    ]



# Generated by Django 3.1.7 on 2021-04-02 20:07

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("data", "0031_auto_20210401_1449"),
    ]

    operations = [
        migrations.AddField(
            model_name="event",
            name="dedupe_key",
            field=models.CharField(max_length=500, null=True),
        ),
    ]



import pytest
import warnings
from datetime import date
from openstates.scrape import Bill
from openstates.utils.generic import get_pseudo_id
from openstates.exceptions import ScrapeValueError


def toy_bill():
    b = Bill(
        identifier="HB 2017",
        legislative_session="2012A",
        title="A bill for an act to raise the cookie budget by 200%",
        from_organization="Foo Senate",
        classification="bill",
    )
    b.add_source("http://uri.example.com/", note="foo")
    return b


def test_basic_valid_bill():
    b = toy_bill()
    b.validate()
    assert "we got here"


def test_bill_type_setting():
    # default
    b = Bill(identifier="some bill", legislative_session="session", title="the title")
    assert b.classification == ["bill"]

    # string -> list
    b = Bill(
        identifier="some bill",
        legislative_session="session",
        title="the title",
        classification="string",
    )
    assert b.classification == ["string"]

    # list unmodified
    b = Bill(
        identifier="some bill",
        legislative_session="session",
        title="the title",
        classification=["two", "items"],
    )
    # will be sorted
    assert b.classification == ["items", "two"]

    # tuple -> list
    b = Bill(
        identifier="some bill",
        legislative_session="session",
        title="the title",
        classification=("two", "items"),
    )
    assert b.classification == ["items", "two"]


def test_basic_invalid_bill():
    """ Test that we can create an invalid bill, and validation will fail """
    b = toy_bill()
    b.identifier = None
    with pytest.raises(ValueError):
        b.validate()


def test_from_organization():
    # none set
    assert get_pseudo_id(Bill("HB 1", "2014", "Some Bill").from_organization) == {
        "classification": "legislature"
    }

    # chamber set
    assert get_pseudo_id(
        Bill("SB 1", "2014", "Some Bill", chamber="upper").from_organization
    ) == {"classification": "upper"}
    # org direct set
    assert (
        Bill("HB 1", "2014", "Some Bill", from_organization="test").from_organization
        == "test"
    )

    # can't set both
    with pytest.raises(ValueError):
        Bill("HB 1", "2014", "Some Bill", from_organization="upper", chamber="upper")


def test_add_action():
    """ Make sure actions work """
    b = toy_bill()
    b.add_action("Some dude liked it.", "2013-04-29T20:00Z", chamber="lower")
    assert len(b.actions) == 1
    assert b.actions[0]["description"] == "Some dude liked it."
    assert get_pseudo_id(b.actions[0]["organization_id"]) == {"classification": "lower"}
    assert b.actions[0]["date"] == "2013-04-29T20:00Z"
    b.validate()


def test_add_related_bill():
    """ Make sure related bills work """
    b = toy_bill()
    b.add_related_bill(
        identifier="HB 2020", legislative_session="2011A", relation_type="companion"
    )
    assert len(b.related_bills) == 1
    assert b.related_bills[0] == {
        "identifier": "HB 2020",
        "legislative_session": "2011A",
        "relation_type": "companion",
    }
    b.validate()


def test_add_sponsor():
    b = toy_bill()
    b.add_sponsorship(
        name="Joe Bleu",
        classification="Author",
        entity_type="person",
        primary=True,
        chamber="upper",
    )
    assert len(b.sponsorships) == 1
    assert b.sponsorships[0] == {
        "person_id": '~{"chamber": "upper", "name": "Joe Bleu"}',
        "name": "Joe Bleu",
        "classification": "Author",
        "entity_type": "person",
        "primary": True,
        "organization_id": None,
    }
    b.validate()


def test_subjects():
    b = toy_bill()
    b.add_subject("Foo")
    b.add_subject("Bar")
    assert b.subject == ["Foo", "Bar"]
    b.validate()


def test_abstract():
    b = toy_bill()
    b.add_abstract("this bill is stupid", "K-5")
    b.add_abstract("this legislative document is ignorant", "6-12")
    assert b.abstracts == [
        {"note": "K-5", "abstract": "this bill is stupid"},
        {"note": "6-12", "abstract": "this legislative document is ignorant"},
    ]


def test_add_documents():
    b = toy_bill()

    # should only add one document since they all have same note
    b.add_document_link(
        note="Fiscal Impact",
        date="2013-04",
        url="http://hi.example.com/foo#bar",
        media_type="text/html",
    )
    b.add_document_link(note="Fiscal Impact", date="2013-04", url="http://foobar.baz")
    assert len(b.documents) == 1

    # should now be two documents
    b.add_document_link(
        note="Other Document", date="2013-04", url="http://foobar.baz/other"
    )
    assert len(b.documents) == 2

    # valid documents so far
    b.validate()

    # an invalid document
    b.add_document_link(
        note="Fiscal Impact", date="2013-04", url=None, media_type="foo"
    )
    with pytest.raises(ScrapeValueError):
        b.validate()


def test_citations():
    b = toy_bill()

    b.add_citation(
        "Wyoming Chapter Laws of 2019",
        "CH0024",
        citation_type="chapter",
        effective=date(2019, 7, 1)
    )

    b.add_citation(
        "Minnesota Session Laws, 2020",
        "Chapter 89",
        citation_type="chapter",
        effective=date(2020, 8, 1),
        url="https://www.revisor.mn.gov/laws/2020/0/Session+Law/Chapter/89/",
    )

    b.add_citation(
        "DC Register",
        "Vol 67 and Page 14429",
        citation_type="final",
        expires="2021-03-06"
    )

    b.add_citation(
        "Constitution of Missouri",
        "Article X Section 6",
        citation_type="proposed",
    )

    b.validate()
    assert len(b.citations) == 4

    with pytest.raises(TypeError):
        # Missing citation
        b.add_citation(
            "Legal Code of Elbonia",
            url="http://openstates.org",
        )

        # Invalid citation
        b.add_citation(
            "Legal Code of Elbonia",
            "",
            citation_type="final",
        )

        # invalid citation type
        b.add_citation(
            "Legal Code of Elbonia",
            "Chapter 12",
            citation_type="aspiration",
        )

        b.validate()


def test_versions():
    b = toy_bill()

    # only one document, multiple links
    b.add_version_link(
        url="http://pault.ag/",
        note="Final Version",
        date="2013-04",
        on_duplicate="error",
        classification="filed",
    )
    b.add_version_link(
        url="http://pault.ag/foo",
        note="Final Version",
        date="2013-04",
        on_duplicate="error",
        classification="filed",
    )
    b.validate()
    assert len(b.versions) == 1
    assert len(b.versions[0]["links"]) == 2

    # duplicate!
    with pytest.raises(ValueError):
        b.add_version_link(
            url="http://pault.ag/foo",
            note="Final Version",
            date="2013-04",
            on_duplicate="error",
        )

    # ignore duplicate - nothing should change
    b.add_version_link(
        url="http://pault.ag/foo",
        note="Final Version",
        date="2013-04",
        classification="filed",
        on_duplicate="ignore",
    )
    assert len(b.versions) == 1
    assert len(b.versions[0]["links"]) == 2

    # duplicate URL
    with pytest.raises(ValueError):
        b.add_version_link(
            url="http://pault.ag/foo",
            note="Finals Versions",
            date="2013-04",
            on_duplicate="error",
        )
    assert len(b.versions) == 1
    assert len(b.versions[0]["links"]) == 2

    # a new doc, numbers go up
    b.add_version_link(
        url="http://pault.ag/foovbar", note="Finals Versions", date="2013-04"
    )
    assert len(b.versions) == 2
    assert len(b.versions[1]["links"]) == 1

    # still validates
    b.validate()


def test_str():
    b = toy_bill()
    assert b.identifier in str(b)


def test_pre_save():
    b = toy_bill()
    b.add_subject("ZZZ")
    b.add_subject("AAA")
    b.add_subject("MMM")
    b.pre_save(None)
    assert b.subject == ["AAA", "MMM", "ZZZ"]


def test_duplicate_sponsor_warning():
    b = toy_bill()
    b.add_sponsorship(
        name="Joe Bleu",
        classification="Author",
        entity_type="person",
        primary=True,
        chamber="upper",
    )
    with warnings.catch_warnings(record=True) as w:
        b.add_sponsorship(
            name="Joe Bleu",
            classification="Author",
            entity_type="person",
            primary=True,
            chamber="upper",
        )
        assert len(w) == 1
        assert "duplicate sponsor" in str(w[0].message)



import re
from django.db.models import Q
from ._types import _JsonDict
from .base import BaseImporter
from ..data.models import Organization


class OrganizationImporter(BaseImporter):
    _type = "organization"
    model_class = Organization

    def limit_spec(self, spec: _JsonDict) -> _JsonDict:
        if spec.get("classification") != "party":
            spec["jurisdiction_id"] = self.jurisdiction_id

        org_name_prepositions = ["and", "at", "by", "for", "in", "on", "of", "the"]
        name = spec.pop("name", None)
        # if chamber is included in pseudo_person_id, we assume this is a committee
        # and chamber is here to help us find its parent
        chamber_classification = spec.pop("chamber", None)
        if name:
            # __icontains doesn't work for JSONField ArrayField
            # so name follows "title" naming pattern
            name = name.title()
            pattern = "(" + "|".join(org_name_prepositions) + ")"
            name = re.sub(
                pattern, lambda match: match.group(0).lower(), name, flags=re.IGNORECASE
            )
            name = name.replace(" & ", " and ")

            if chamber_classification:
                return (
                    Q(**spec)
                    & (Q(name__iexact=name) | Q(other_names__contains=[{"name": name}]))
                    & Q(parent__classification=chamber_classification)
                )
            else:
                return Q(**spec) & (
                    Q(name__iexact=name) | Q(other_names__contains=[{"name": name}])
                )
        return spec
